@@ -1,3 +1,4 @@
+import copy
 import logging
 from contextlib import contextmanager
 from functools import lru_cache
@@ -7,13 +8,24 @@ from requests import Session
 from requests.exceptions import HTTPError
 
 from .. import query as q
-from ..utils import _camel_to_snake_case, _sanitize_session_id
+from ..__about__ import __version__
+from ..systems import (
+    ELMO_E_CONNECT,
+    ELMO_E_CONNECT_WEB_LOGIN,
+    IESS_METRONET,
+    IESS_METRONET_WEB_LOGIN,
+)
+from ..utils import (
+    _camel_to_snake_case,
+    _sanitize_session_id,
+    extract_session_id_from_html,
+)
 from .decorators import require_lock, require_session
 from .exceptions import (
     CodeError,
+    CommandError,
     CredentialError,
-    InvalidInput,
-    InvalidSector,
+    DeviceDisconnectedError,
     InvalidToken,
     LockError,
     ParseError,
@@ -45,14 +57,30 @@ class ElmoClient:
         self._domain = domain
         self._session = Session()
         self._session_id = session_id
+        self._panel = None
         self._lock = Lock()
+
+        # Web login is required for E-Connect and Metronet because, at the moment, the
+        # Cloud API login does not register the client session in their backend, causing
+        # updates to be not received by the client during the long-polling `has_updates`.
+        domain_url = "" if self._domain == "default" else self._domain
+        self._web_login = base_url in [ELMO_E_CONNECT, IESS_METRONET]
+        if base_url == ELMO_E_CONNECT:
+            self._web_login_url = f"{ELMO_E_CONNECT_WEB_LOGIN}/{domain_url}"
+        elif base_url == IESS_METRONET:
+            self._web_login_url = f"{IESS_METRONET_WEB_LOGIN}/{domain_url}"
+        else:
+            self._web_login_url = None
+
         # Debug
+        _LOGGER.debug(f"Client | Library version: {__version__}")
         _LOGGER.debug(f"Client | Router: {self._router._base_url}")
         _LOGGER.debug(f"Client | Domain: {self._domain}")
+        _LOGGER.debug(f"Client | Web login URL: {self._web_login_url}")
 
     def auth(self, username, password):
         """Authenticate the client and retrieves the access token. This method uses
-        the Authentication API.
+        the Authentication API, or the web login form if the base_url is Elmo E-Connect.
 
         Args:
             username: the Username used for the authentication.
@@ -69,7 +97,7 @@ class ElmoClient:
             if self._domain is not None:
                 payload["domain"] = self._domain
 
-            _LOGGER.debug("Client | Client Authentication")
+            _LOGGER.debug("Client | API Authentication")
             response = self._session.get(self._router.auth, params=payload)
             response.raise_for_status()
         except HTTPError as err:
@@ -78,9 +106,10 @@ class ElmoClient:
                 raise CredentialError
             raise err
 
-        # Store the session_id
+        # Store the session_id and the panel details (if available)
         data = response.json()
         self._session_id = data["SessionId"]
+        self._panel = {_camel_to_snake_case(k): v for k, v in data.get("Panel", {}).items()}
 
         # Register the redirect URL and try the authentication again
         if data["Redirect"]:
@@ -91,6 +120,19 @@ class ElmoClient:
             data = redirect.json()
             self._session_id = data["SessionId"]
 
+        # Retrieve the session_id using the web login form
+        if self._web_login:
+            payload = {
+                "IsDisableAccountCreation": "True",
+                "IsAllowThemeChange": "True",
+                "UserName": username,
+                "Password": password,
+                "RememberMe": "false",
+            }
+            web_response = self._session.post(self._web_login_url, data=payload)
+            web_response.raise_for_status()
+            self._session_id = extract_session_id_from_html(web_response.text)
+
         _LOGGER.debug(f"Client | Authentication successful: {_sanitize_session_id(self._session_id)}")
         return self._session_id
 
@@ -151,13 +193,15 @@ class ElmoClient:
 
     @contextmanager
     @require_session
-    def lock(self, code):
+    def lock(self, code, user_id=1):
         """Context manager to obtain a system lock. The alerting system allows
         only one user at a time and obtaining the lock is mandatory. When the
         context manager is closed, the lock is automatically released.
 
         Args:
             code: the alarm code used to obtain the lock.
+            user_id: the `userId` used by some main units. This value is optional and
+                should be used only if the main unit requires it. The default value is 1.
         Raises:
             CodeError: if used `code` is not valid.
             LockError: if the server is refusing to assign the lock. It could mean
@@ -167,8 +211,10 @@ class ElmoClient:
         Returns:
             A client instance with an acquired lock.
         """
-        payload = {"userId": 1, "password": code, "sessionId": self._session_id}
+        # Main units that do not require a userId param, expects userId to be "1"
+        payload = {"userId": user_id, "password": code, "sessionId": self._session_id}
         response = self._session.post(self._router.lock, data=payload)
+        _LOGGER.debug(f"Client | Lock response: {response.text}")
 
         try:
             response.raise_for_status()
@@ -187,7 +233,7 @@ class ElmoClient:
             raise CodeError
 
         self._lock.acquire()
-        _LOGGER.debug(f"Client | Lock acquired with response {body}")
+        _LOGGER.debug("Client | Lock successful")
         try:
             yield self
         finally:
@@ -211,9 +257,10 @@ class ElmoClient:
         """
         payload = {"sessionId": self._session_id}
         response = self._session.post(self._router.unlock, data=payload)
+        _LOGGER.debug(f"Client | Unlock response: {response.text}")
         response.raise_for_status()
 
-        _LOGGER.debug(f"Client | Lock released with response {response.text}")
+        _LOGGER.debug("Client | Unlock successful")
         # Release the lock only in case of success, so that if it fails
         # the owner of the lock can properly unlock the system again
         # (maybe with a retry)
@@ -262,21 +309,15 @@ class ElmoClient:
 
         # Send the payload to arm sectors
         response = self._session.post(self._router.send_command, data=payload)
+        _LOGGER.debug(f"Client | Arm response: {response.text}")
         response.raise_for_status()
-
-        errors = []
-        # A not existing sector returns 200 with a fail state
         body = response.json()
-        _LOGGER.debug(f"Client | Arming response: {body}")
+
+        # Errors returns 200 with "Successful == False" JSON key
         if not body[0]["Successful"]:
-            errors.append(payload["ElementsIndexes"])
+            raise CommandError
 
-        # Raise an exception if errors are detected
-        if errors:
-            invalid_sectors = ",".join(str(x) for x in errors)
-            raise InvalidSector("Selected sectors don't exist: {}".format(invalid_sectors))
-
-        _LOGGER.debug("Client | Arming successful")
+        _LOGGER.debug("Client | Arm successful")
         return True
 
     @require_session
@@ -322,21 +363,15 @@ class ElmoClient:
 
         # Send the payload to disarm sectors
         response = self._session.post(self._router.send_command, data=payload)
+        _LOGGER.debug(f"Client | Disarm response: {response.text}")
         response.raise_for_status()
-
-        errors = []
-        # A not existing sector returns 200 with a fail state
         body = response.json()
-        _LOGGER.debug(f"Client | Disarming response: {body}")
+
+        # Errors returns 200 with "Successful == False" JSON key
         if not body[0]["Successful"]:
-            errors.append(payload["ElementsIndexes"])
+            raise CommandError
 
-        # Raise an exception if errors are detected
-        if errors:
-            invalid_sectors = ",".join(str(x) for x in errors)
-            raise InvalidSector("Selected sectors don't exist: {}".format(invalid_sectors))
-
-        _LOGGER.debug("Client | Disarming successful")
+        _LOGGER.debug("Client | Disarm successful")
         return True
 
     @require_session
@@ -392,7 +427,7 @@ class ElmoClient:
         # Raise an exception if errors are detected
         if errors:
             invalid_inputs = ",".join(str(x) for x in errors)
-            raise InvalidInput("Selected inputs don't exist: {}".format(invalid_inputs))
+            raise CommandError("Selected inputs don't exist: {}".format(invalid_inputs))
 
         _LOGGER.debug("Client | Excluding successful")
         return True
@@ -450,7 +485,7 @@ class ElmoClient:
         # Raise an exception if errors are detected
         if errors:
             invalid_inputs = ",".join(str(x) for x in errors)
-            raise InvalidInput("Selected inputs don't exist: {}".format(invalid_inputs))
+            raise CommandError("Selected inputs don't exist: {}".format(invalid_inputs))
 
         _LOGGER.debug("Client | Including successful")
         return True
@@ -488,24 +523,17 @@ class ElmoClient:
             "sessionId": self._session_id,
         }
 
-        errors = []
-
         # Send turn on request
         response = self._session.post(self._router.send_command, data=payload)
         response.raise_for_status()
-
-        # A not existing output returns 200 with a fail state
         body = response.json()
-        _LOGGER.debug(f"Client | Turning on response: {body}")
+
+        # Errors returns 200 with "Successful == False" JSON key
         if not body[0]["Successful"]:
-            errors.append(payload["ElementsIndexes"])
+            _LOGGER.error(f"Client | Turning on response: {body}")
+            raise CommandError
 
-        # Raise an exception if errors are detected
-        if errors:
-            invalid_outputs = ",".join(str(x) for x in errors)
-            raise InvalidInput("Selected outputs don't exist: {}".format(invalid_outputs))
-
-        _LOGGER.debug("Client | Turning on output successful")
+        _LOGGER.debug(f"Client | Turning on successful with response: {body}")
         return True
 
     @require_session
@@ -544,20 +572,14 @@ class ElmoClient:
         # Send turn off request
         response = self._session.post(self._router.send_command, data=payload)
         response.raise_for_status()
-
-        errors = []
-        # A not existing output returns 200 with a fail state
         body = response.json()
-        _LOGGER.debug(f"Client | Turning off response: {body}")
+
+        # Errors returns 200 with "Successful == False" JSON key
         if not body[0]["Successful"]:
-            errors.append(payload["ElementsIndexes"])
+            _LOGGER.error(f"Client | Turning on response: {body}")
+            raise CommandError
 
-        # Raise an exception if errors are detected
-        if errors:
-            invalid_outputs = ",".join(str(x) for x in errors)
-            raise InvalidInput("Selected outputs don't exist: {}".format(invalid_outputs))
-
-        _LOGGER.debug("Client | Turning off output successful")
+        _LOGGER.debug(f"Client | Turning on successful with response: {body}")
         return True
 
     @lru_cache(maxsize=1)
@@ -628,12 +650,24 @@ class ElmoClient:
         elif query == q.ALERTS:
             endpoint = self._router.status
             _LOGGER.debug("Client | Querying alerts")
+        elif query == q.PANEL:
+            _LOGGER.debug("Client | Querying panel details (cached)")
+            return {
+                "last_id": 0,
+                "panel": copy.deepcopy(self._panel) if self._panel else {},
+            }
         else:
             # Bail-out if the query is not recognized
             raise QueryNotValid()
 
-        response = self._session.post(endpoint, data={"sessionId": self._session_id})
-        response.raise_for_status()
+        try:
+            response = self._session.post(endpoint, data={"sessionId": self._session_id})
+            response.raise_for_status()
+        except HTTPError as err:
+            # Handle the case when the device is disconnected
+            if err.response.status_code == 403 and "Centrale non connessa" in err.response.text:
+                raise DeviceDisconnectedError
+            raise err
 
         if query in [q.SECTORS, q.INPUTS, q.OUTPUTS]:
             # Retrieve description or use the cache
@@ -646,8 +680,17 @@ class ElmoClient:
             entries = response.json()
             _LOGGER.debug(f"Client | Query response: {entries}")
             items = {}
+
+            try:
+                # Determine the last_id by finding the maximum Id in the list
+                last_id = max(entry["Id"] for entry in entries)
+            except (TypeError, ValueError) as err:
+                _LOGGER.error("Client | Could not determine max Id from entries, defaulting to 0.")
+                _LOGGER.debug(f"Client | Error: {err} | Entries: {entries}")
+                last_id = 0
+
             result = {
-                "last_id": entries[-1]["Id"],
+                "last_id": last_id,
                 key_group: items,
             }
             try:
