@@ -239,6 +239,86 @@ def test_client_auth_redirect(server):
     assert len(server.calls) == 2
 
 
+def test_client_auth_redirect_web_login_econnect(server):
+    """Ensure web login session token is used when redirect is required.
+    Regression test: https://github.com/palazzem/econnect-python/issues/158
+    """
+    redirect = """
+        {
+            "SessionId": "00000000-0000-0000-0000-000000000000",
+            "Domain": "domain",
+            "Redirect": true,
+            "RedirectTo": "https://redirect.example.com"
+        }
+    """
+    login = """
+        {
+            "SessionId": "99999999-9999-9999-9999-999999999999",
+            "Username": "test",
+            "Domain": "domain",
+            "Language": "en",
+            "IsActivated": true,
+            "IsConnected": true,
+            "IsLoggedIn": false,
+            "IsLoginInProgress": false,
+            "CanElevate": true,
+            "AccountId": 100,
+            "IsManaged": false,
+            "Redirect": false,
+            "IsElevation": false
+        }
+    """
+    server.add(responses.GET, "https://connect.elmospa.com/api/login", body=redirect, status=200)
+    server.add(responses.GET, "https://redirect.example.com/api/login", body=login, status=200)
+    server.add(responses.POST, "https://webservice.elmospa.com/domain", body=r.STATUS_PAGE, status=200)
+    client = ElmoClient(base_url=ELMO_E_CONNECT, domain="domain")
+    # Test
+    assert client.auth("test", "test")
+    assert len(server.calls) == 3
+    assert client._router._base_url == "https://redirect.example.com"
+    assert client._session_id == "f8h23b4e-7a9f-4d3f-9b08-2769263ee33c"
+
+
+def test_client_auth_redirect_web_login_metronet(server):
+    """Ensure web login session token is used when redirect is required.
+    Regression test: https://github.com/palazzem/econnect-python/issues/168
+    """
+    redirect = """
+        {
+            "SessionId": "00000000-0000-0000-0000-000000000000",
+            "Domain": "domain",
+            "Redirect": true,
+            "RedirectTo": "https://redirect.example.com"
+        }
+    """
+    login = """
+        {
+            "SessionId": "99999999-9999-9999-9999-999999999999",
+            "Username": "test",
+            "Domain": "domain",
+            "Language": "en",
+            "IsActivated": true,
+            "IsConnected": true,
+            "IsLoggedIn": false,
+            "IsLoginInProgress": false,
+            "CanElevate": true,
+            "AccountId": 100,
+            "IsManaged": false,
+            "Redirect": false,
+            "IsElevation": false
+        }
+    """
+    server.add(responses.GET, "https://metronet.iessonline.com/api/login", body=redirect, status=200)
+    server.add(responses.GET, "https://redirect.example.com/api/login", body=login, status=200)
+    server.add(responses.POST, "https://metronet.iessonline.com/domain", body=r.STATUS_PAGE, status=200)
+    client = ElmoClient(base_url=IESS_METRONET, domain="domain")
+    # Test
+    assert client.auth("test", "test")
+    assert len(server.calls) == 3
+    assert client._router._base_url == "https://redirect.example.com"
+    assert client._session_id == "f8h23b4e-7a9f-4d3f-9b08-2769263ee33c"
+
+
 def test_client_auth_infinite_redirect(server):
     """Should prevent infinite redirects in the auth() call."""
     redirect = """
@@ -348,6 +428,90 @@ def test_client_poll(server):
     assert "ConnectionStatus=1" in body
 
 
+def test_client_auth_econnect_web_login(server):
+    """Ensure API and Web login are executed when using e-Connect cloud API.
+    Regression test: https://github.com/palazzem/econnect-python/issues/158
+    """
+    server.add(responses.GET, "https://connect.elmospa.com/api/login", body=r.LOGIN, status=200)
+    server.add(responses.POST, "https://webservice.elmospa.com/domain", body=r.STATUS_PAGE, status=200)
+    client = ElmoClient(base_url=ELMO_E_CONNECT, domain="domain")
+    # Test
+    client.auth("test", "test")
+    assert len(server.calls) == 2
+    request_body = dict(item.split("=") for item in server.calls[1].request.body.split("&"))
+    assert client._session_id == "f8h23b4e-7a9f-4d3f-9b08-2769263ee33c"
+    assert request_body == {
+        "IsDisableAccountCreation": "True",
+        "IsAllowThemeChange": "True",
+        "UserName": "test",
+        "Password": "test",
+        "RememberMe": "false",
+    }
+
+
+def test_client_auth_econnect_web_login_with_default_domain(server):
+    """Ensure API and Web login are executed when using e-Connect cloud API.
+    Regression test: https://github.com/palazzem/econnect-python/issues/158
+    """
+    server.add(responses.GET, "https://connect.elmospa.com/api/login", body=r.LOGIN, status=200)
+    server.add(responses.POST, "https://webservice.elmospa.com/", body=r.STATUS_PAGE, status=200)
+    client = ElmoClient(base_url=ELMO_E_CONNECT, domain="default")
+    # Test
+    client.auth("test", "test")
+    assert len(server.calls) == 2
+    request_body = dict(item.split("=") for item in server.calls[1].request.body.split("&"))
+    assert client._session_id == "f8h23b4e-7a9f-4d3f-9b08-2769263ee33c"
+    assert request_body == {
+        "IsDisableAccountCreation": "True",
+        "IsAllowThemeChange": "True",
+        "UserName": "test",
+        "Password": "test",
+        "RememberMe": "false",
+    }
+
+
+def test_client_auth_econnect_web_login_metronet(server):
+    """Web login must be used when accessing with Metronet.
+    Regression test: https://github.com/palazzem/econnect-python/issues/186
+    """
+    server.add(responses.GET, "https://metronet.iessonline.com/api/login", body=r.LOGIN, status=200)
+    server.add(responses.POST, "https://metronet.iessonline.com/domain", body=r.STATUS_PAGE, status=200)
+    client = ElmoClient(base_url=IESS_METRONET, domain="domain")
+    # Test
+    client.auth("test", "test")
+    assert len(server.calls) == 2
+    request_body = dict(item.split("=") for item in server.calls[1].request.body.split("&"))
+    assert client._session_id == "f8h23b4e-7a9f-4d3f-9b08-2769263ee33c"
+    assert request_body == {
+        "IsDisableAccountCreation": "True",
+        "IsAllowThemeChange": "True",
+        "UserName": "test",
+        "Password": "test",
+        "RememberMe": "false",
+    }
+
+
+def test_client_auth_metronet_web_login_with_default_domain(server):
+    """Web login must be used when accessing with Metronet.
+    Regression test: https://github.com/palazzem/econnect-python/issues/186
+    """
+    server.add(responses.GET, "https://metronet.iessonline.com/api/login", body=r.LOGIN, status=200)
+    server.add(responses.POST, "https://metronet.iessonline.com/", body=r.STATUS_PAGE, status=200)
+    client = ElmoClient(base_url=IESS_METRONET, domain="default")
+    # Test
+    client.auth("test", "test")
+    assert len(server.calls) == 2
+    request_body = dict(item.split("=") for item in server.calls[1].request.body.split("&"))
+    assert client._session_id == "f8h23b4e-7a9f-4d3f-9b08-2769263ee33c"
+    assert request_body == {
+        "IsDisableAccountCreation": "True",
+        "IsAllowThemeChange": "True",
+        "UserName": "test",
+        "Password": "test",
+        "RememberMe": "false",
+    }
+
+
 def test_client_poll_with_changes(server):
     """Should return a dict with updated states."""
     html = """
@@ -2562,3 +2726,98 @@ def test_client_get_alerts_missing_data(server):
     with pytest.raises(ParseError):
         client.query(query.ALERTS)
     assert len(server.calls) == 1
+
+
+def test_client_query_last_id_unordered(server, mocker):
+    """Should determine the last_id correctly even if entries are unordered.
+    Regression test for: https://github.com/palazzem/econnect-python/issues/154
+    """
+    html = """[
+       {
+           "Alarm": true,
+           "MemoryAlarm": false,
+           "Excluded": false,
+           "InUse": true,
+           "IsVideo": false,
+           "Id": 3,
+           "Index": 0,
+           "Element": 1,
+           "CommandId": 0,
+           "InProgress": false
+       },
+       {
+           "Alarm": false,
+           "MemoryAlarm": false,
+           "Excluded": true,
+           "InUse": true,
+           "IsVideo": false,
+           "Id": 5,
+           "Index": 2,
+           "Element": 3,
+           "CommandId": 0,
+           "InProgress": false
+       },
+       {
+           "Alarm": true,
+           "MemoryAlarm": false,
+           "Excluded": false,
+           "InUse": true,
+           "IsVideo": false,
+           "Id": 2,
+           "Index": 1,
+           "Element": 2,
+           "CommandId": 0,
+           "InProgress": false
+       }
+    ]"""
+    server.add(responses.POST, "https://example.com/api/inputs", body=html, status=200)
+    client = ElmoClient(base_url="https://example.com", domain="domain")
+    client._session_id = "test"
+    mocker.patch.object(client, "_get_descriptions")
+    client._get_descriptions.return_value = {
+        10: {0: "Input 1", 1: "Input 2", 2: "Input 3"},
+    }
+    # Test
+    inputs = client.query(query.INPUTS)
+    assert inputs["last_id"] == 5
+
+
+def test_client_query_last_id_type_error(server, mocker):
+    """Should default last_id to 0 if a TypeError occurs during max()."""
+    html = """[
+       {
+           "Alarm": true,
+           "InUse": true,
+           "Id": "not-an-int",
+           "Index": 0,
+           "Element": 1
+       },
+       {
+           "Alarm": false,
+           "InUse": true,
+           "Id": 2,
+           "Index": 1,
+           "Element": 2
+       }
+    ]"""
+    server.add(responses.POST, "https://example.com/api/inputs", body=html, status=200)
+    client = ElmoClient(base_url="https://example.com", domain="domain")
+    client._session_id = "test"
+    mocker.patch.object(client, "_get_descriptions")
+    client._get_descriptions.return_value = {10: {0: "Input 1", 1: "Input 2"}}
+    # Test
+    inputs = client.query(query.INPUTS)
+    assert inputs["last_id"] == 0
+
+
+def test_client_query_last_id_value_error_empty(server, mocker):
+    """Should default last_id to 0 if the entries list is empty (ValueError)."""
+    html = """[]"""
+    server.add(responses.POST, "https://example.com/api/inputs", body=html, status=200)
+    client = ElmoClient(base_url="https://example.com", domain="domain")
+    client._session_id = "test"
+    mocker.patch.object(client, "_get_descriptions")
+    client._get_descriptions.return_value = {}
+    # Test
+    inputs = client.query(query.INPUTS)
+    assert inputs["last_id"] == 0
